from typing import Sequence, Optional

import canopy
import pandas as pd
import json

_definition_key = 'definition'
_sim_types_key = 'simTypes'
_sim_config_key = 'simConfig'


class StudyResult:
    _sim_types: Optional[Sequence[str]] = None
    _inputs: Optional[canopy.DynamicDictToObject] = None

    def __init__(
            self,
            session: canopy.Session,
            study_result: Optional[canopy.openapi.GetStudyQueryResult],
            jobs: Optional[Sequence[canopy.StudyJobResult]],
            document: Optional[canopy.openapi.CanopyDocument] = None,
            scalar_results: Optional[canopy.StudyScalarResults] = None):
        self._session = session
        self._study_result = study_result
        self._scalar_results = scalar_results
        self._document: canopy.openapi.CanopyDocument = document if document is not None else study_result.study

        if self._document is None:
            raise RuntimeError('Study document not provided.')

        self._data = canopy.get_study_document(self._session, self._document)
        self._jobs = jobs

        if self._data is not None and self._data.definition is not None:
            definition = canopy.ensure_dict(self._data.definition)
            self._sim_types = definition[_sim_types_key]
            self._inputs = canopy.DynamicDictToObject(canopy.ensure_dict(definition[_sim_config_key]))

    @property
    def result(self) -> canopy.openapi.GetStudyQueryResult:
        return self._study_result

    @property
    def document(self) -> canopy.openapi.CanopyDocument:
        return self._document

    @property
    def study_id(self) -> str:
        return self._document.document_id

    @property
    def data(self) -> canopy.openapi.StudyDocument:
        return self._data

    @property
    def sim_types(self) -> Optional[Sequence[str]]:
        return self._sim_types

    @property
    def inputs(self) -> Optional[canopy.DynamicDictToObject]:
        return self._inputs

    @property
    def jobs(self) -> Optional[Sequence[canopy.StudyJobResult]]:
        return self._jobs

    @property
    def scalar_results(self) -> Optional[canopy.StudyScalarResults]:
        return self._scalar_results

    @property
    def simulation_count(self) -> int:
        if self._data.job_count == 1:
            return 1

        # If there is more than one job then the final job will be the post-processing task.
        return self._data.job_count - 1

    @property
    def succeeded_simulation_count(self) -> int:
        return self._data.succeeded_simulation_count

    def scalar_as(self, channel_name: str, target_unit: Optional[str] = None, always_return_copy: bool = False) -> Optional[pd.Series]:
        if self._scalar_results is None:
            return None

        scalar_results = self._scalar_results
        if target_unit is None:
            target_unit = self._session.user_settings.get_channel_units(channel_name)

        result: Optional[pd.Series] = None
        if scalar_results.inputs is not None and channel_name in scalar_results.inputs.columns:
            result = scalar_results.inputs[channel_name]
        elif scalar_results.results is not None and channel_name in scalar_results.results.columns:
            result = scalar_results.results[channel_name]

        if result is None:
            return None

        if target_unit is None:
            return result.copy() if always_return_copy else result

        source_unit = ''
        if channel_name in self._scalar_results.units:
            source_unit = self._scalar_results.units[channel_name]

        return self._session.units.convert_series_between_units(
            result,
            source_unit,
            target_unit,
            inplace=False,
            always_return_copy=always_return_copy)

    def to_dict(self):
        return {
            'study_id': self.study_id,
            'sim_types': self._sim_types,
            'simulation_count': self.simulation_count,
            'study_result': self._study_result,
            'document': self.document
        }

    def __repr__(self):
        return 'canopy.StudyResult(%r,%r,%r)' % \
               (self.study_id, self._sim_types, self.simulation_count)

    def __str__(self):
        return json.dumps(self.to_dict(), indent=2, default=str)


from typing import Optional

import os
import canopy
import integration_tests
import uuid
import logging

logger = logging.getLogger(__name__)

_python_data_version = '1'


class EnvironmentLoader:
    _data: Optional[integration_tests.EnvironmentData] = None

    @staticmethod
    async def load() -> integration_tests.EnvironmentData:
        if EnvironmentLoader._data is None:
            credentials_key = 'CANOPY_PYTHON_INTEGRATION_TEST_CREDENTIALS'
            credentials = os.environ.get(credentials_key)

            if credentials is None:
                raise RuntimeError(f'Test credentials not found. Please create an environment variable "{credentials_key}"="client_id|client_secret|user_name|tenant_name|password"')

            client_id, client_secret, username, tenant_name, password = credentials.split('|')

            if client_id is None:
                raise RuntimeError('Client ID not found.')
            if client_secret is None:
                raise RuntimeError('Client Secret not found.')
            if username is None:
                raise RuntimeError('Username not found.')
            if tenant_name is None:
                raise RuntimeError('Tenant Name not found.')
            if password is None:
                raise RuntimeError('Password not found.')

            authentication_data = canopy.AuthenticationData(
                client_id=client_id,
                client_secret=client_secret,
                username=username,
                tenant_name=tenant_name,
                password=password)

            if username == 'master' and tenant_name == 'test':
                authentication_data = await EnvironmentLoader.create_test_user(authentication_data)

            EnvironmentLoader._data = integration_tests.EnvironmentData(
                authentication_data)

        return EnvironmentLoader._data

    @staticmethod
    async def create_test_user(master: canopy.AuthenticationData) -> canopy.AuthenticationData:
        username = 'u-' + str(uuid.uuid4()).replace('-', '')[0:18]
        password = str(uuid.uuid4())
        async with canopy.Session(authentication_data=master) as session:
            session.authentication.authenticate()

            api = canopy.openapi.MembershipApi(session.async_client)
            await api.membership_post_registration(
                canopy.openapi.RegistrationData(
                    session.authentication.tenant_id,
                    username + '@testing.canopysimulations.com',
                    username,
                    password))

            logger.info('Created test user: {}'.format(username))

            return canopy.AuthenticationData(
                master.client_id,
                master.client_secret,
                username,
                master.tenant_name,
                password)



from typing import Optional, List, Dict

import canopy


async def find_study(
        session: canopy.Session,
        name: Optional[str] = None,
        owner_username: Optional[str] = None,
        custom_properties: Dict[str, str] = None,
        parent_worksheet_id: Optional[str] = None,
        tenant_id: Optional[str] = None) -> canopy.StudyResult:

    session.authentication.authenticate()

    if tenant_id is None:
        tenant_id = session.authentication.tenant_id

    study_api = canopy.openapi.StudyApi(session.async_client)

    filter_ = canopy.create_list_filter(
        session,
        is_study=True,
        name=name,
        items_per_page=1,
        owner_username=owner_username,
        custom_properties=custom_properties,
        parent_worksheet_id=parent_worksheet_id,
        tenant_id=tenant_id)

    metadata_result: canopy.openapi.GetConfigsQueryResult = await study_api.study_get_studies(
        tenant_id,
        **canopy.defined_kwargs(
            filter=filter_.serialize()))

    documents: List[canopy.openapi.CanopyDocument] = metadata_result.query_results.documents
    if len(documents) == 0:
        raise canopy.NotFoundError('No study found matching the specified criteria.')

    return canopy.StudyResult(
        session,
        None,
        None,
        document=documents[0])



from typing import Sequence, Optional

import canopy


class TenantUsers:
    def __init__(
            self,
            result: canopy.openapi.GetTenantUsersQueryResult):
        self._list = result.users
        self._by_user_id = {item.user_id: item for item in result.users}
        self._by_username = {item.username: item for item in result.users}

    @property
    def list(self) -> Sequence[canopy.openapi.GetTenantUsersQueryResultUserItem]:
        return self._list

    def get_by_user_id(self, user_id: str) -> Optional[canopy.openapi.GetTenantUsersQueryResultUserItem]:
        if user_id in self._by_user_id:
            return self._by_user_id[user_id]
        else:
            raise canopy.NotFoundError('Unknown user ID ' + user_id)

    def get_by_username(self, username: str) -> Optional[canopy.openapi.GetTenantUsersQueryResultUserItem]:
        if username in self._by_username:
            return self._by_username[username]
        else:
            raise canopy.NotFoundError('Unknown username ' + username)
